/*
  GPL Driver for the Codec Interface on MGA G1/2/400 cards
  (requires mgacore.c)

  (c) 1999-2000 David Barth <dbarth@besancon.net>
*/

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include "mgacore.h"
#include "mgacodec.h"

/* TODO :
   - implement mgacodec_read_bitstream
   - iovec inputs ?
*/

/*
 * define this macro to enable code specific to the DVD driver
 */
#define MGADVD_HACK

/*
 * define this macro to enable code specific to the ZR36060 driver
 */
/* #define ZR36060_HACK */

/****************************************************************************
 * MATROX CODEC INTERFACE                                                   *
 ****************************************************************************/

/****************************************************************************
 * PRIVATE SECTION : implementations details                                *
 *                                                                          *
 * (see the end of the file for public functions implementation)            *
 *                                                                          *
 ****************************************************************************/

#define CODEC_NO_TRANSFER	0
#define CODEC_SEND_TRANSFER	1
#define CODEC_RECV_TRANSFER	2

/*
 * Private structure definition. Only the codec module has to know the
 * details. Modules using the codec, should access it through the
 * public API.
 */
   
struct mga_codec_s
{
	/* address of the codec interface inside the framebuffer */
        u32 addr;

	/* codec interface mode (VMI or I33) */
	int mode;

	/* VMI mode, 0 = A, 1 = B */
	int vmi;

	/* register (in the external chip) to send bitstream to */
	int dstreg;

	/* recovery width, see G200 spec, p 4-89 for timing information */
	int rwidth;

	/* shadow registers for CODECCTL:
	 * these are Write-Only registers on the G200, so this copy lets
	 * us read it to see what the latest settings are
	 * NOTE: CODECCTL+2, bits 16-23 of CODECCTL are unused.
	 * WARNING: CODECCTL register should ONLY be programmed byte at a time
	 * -inkling
	 */
	u8 misc_shadow;		/* CODECCTL+3, miscctl byte */
	u8 rwcfa_shadow;	/* CODECCTL+1, recovery width, codecfifoaddr */
	u8 ctl_shadow;		/* CODECCTL+0, general codec control bits */
	
	/* buffers in host memory holding a copy of the command and data
	   buffer of the codec interface.  The command buffer is copied
	   at execution time to the framebuffer. Once the commands have
	   been executed the databuffer is filled with the data retrieved
	   from the codec and waiting in the framebuffer.
	*/
	u32 cmdbuffer[64]; /* organized as groups of 8 bytes */
	int cmdindex;
	u16 databuffer[128]; /* organized as groups of 8 bytes */
	int dataindex;

	/* see mgacodec_send_bitstream */
	u32 streamindex;

	/*
	 * transfer indicates a codec transfer has started and also which 
	 * direction it is going. defined values are (numeric):
	 *  CODEC_NO_TRANSFER	(0)	no transfer has started yet
	 *  CODEC_SEND_TRANSFER	(1)	data FROM G200 to the codec
	 *  CODEC_RECV_TRANSFER	(2)	data TO G200 from the codec
	 *
	 * stalled indicates a buffer level IRQ has stalled the transfer
	 * 
	 */
	int transfer;
	int stalled;

	wait_queue_head_t outq;
	wait_queue_head_t inq;
	wait_queue_head_t cmdq; /* not used yet */

	struct mga_dev_s *mga;
};

static struct mga_codec_s unique_codec;

/* forward declaration */
static void mgacodec_buflevel_irq(struct mga_dev_s *mga, struct pt_regs *regs);

/* CODEC registers */
#define CODECCTL        0x3e40
#define CODECMISCCTL	0x3e43
#define CODECADDR       0x3e44
#define CODECHOSTPTR    0x3e48
#define CODECHARDPTR    0x3e4c
#define CODECLCODE      0x3e50

/*
 * CODECCTL fields and commands
 */
#define codecen 	0x0001	/* CODEC enable, invert to disable */
#define codecmode 	0x0002	/* CODEC I33 mode, invert for VMI mode */
#define cmdexectrig 	0x0004	/* command execution trigger, invert to prevent execution */
#define codecdatain 	0x0008	/* compression, invert for decompression */
#define vmimode 	0x0010	/* VMI mode B, invert for mode A */
#define stopcodec 	0x0020	/* stop after current field, invert to not stop */
#define codectransen 	0x0040	/* enable transfer, invert to disable */

/*
 * VSTATUS bits
	not all of these are used by the 36710 (mgadvd)
	i've indicated the ones that are with ++ and the ones that should
	be with +-. other modules, i.e. RRG, will use a different subset
	see G200 spec, p 3-198 
	-inkling
 */
#define vinvsyncpen	0x0001	/*    video in vertical sync pending */
#define cmdcmplpen	0x0002	/* +- command complete interrupt pending */
#define blvlpen		0x0004	/* ++ buffer level interrupt pending */
#define dcmpeoipen	0x0008	/*    decomp end of image interrupt pending */
#define vinfielddetd	0x0100	/*    video input field detected */
#define vincapd		0x0200	/*    video in captured */
#define rawvbicapd	0x0400	/*    raw vbi captured */
#define slcvbicapd	0x0800	/*    slice vbi captured */
#define codec_stalled 	0x1000	/* ++ special use conditions */
/*
  codec_stalled is ONLY valid under certain conditions. see G200 spec p 3-199
  -inkling
*/

/* re/set codecctl bits in copy and use copy to set codecctl */
static __inline__ void mgacodec_and_or_codecctl(struct mga_codec_s *codec, u8 and, u8 or)
{
	codec->ctl_shadow &= and;
	codec->ctl_shadow |= or;

	/* write byte NOT word because according to the G200 spec, p 4-89,
	   writing to codec recovery/codecfifoaddr byte can cause corruption
	   during data transfer. this agress with the working april code.
	   see mgacodec_reset for setting the recovery width at CODECCTL+1
	   -inkling */

	mga_writeb(codec->mga, CODECCTL, codec->ctl_shadow);
//	printk("mgacodec: ctl_mask 0x%02X ctl 0x%02X shadow 0x%02X\n",
//		and, or, codec->ctl_shadow);
}

#define CODEC_WRITE(addr, byte) ( ((addr) << 8) | (byte) )

/*
 * wait some microseconds
 *
 */
static inline void mgacodec_wait(int delay)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(delay * HZ/1000000);
}

/* puts the codec interface in bitstream transfer mode
   sets transfer flag, clears stalled flag, and sets direction
*/
static int mgacodec_start_transfer(struct mga_codec_s *codec, int direction)
{
	u8 ctl_mask = codec->ctl_shadow;
	u8 ctl = codecen | codectransen;

	/* codec recovery width, play with this number to see what you get.
	   i did not notice any difference no matter what setting i used.
	   it may be more useful for the other encoder/decoders, and thus
	   should be part of mgacodec_init.
	   - inkling */
	codec->rwidth = 1;

	/* build the byte to send to CODECCTL + 1, with sanity limits */
	codec->rwcfa_shadow = ((codec->rwidth & 3)<<4) | (codec->dstreg & 0xF) ;

	if (codec->transfer)
	{
	/* check if we need to restart a transfer that has been stalled
	   because of a codec buffer level IRQ occurring
	   -inkling */
		if (codec->stalled)
		{
			/* keep current ctl (codectransen should be 0) */
			codec->stalled = 0;
			/*
			printk("mgacodec: %s transfer re-starting\n",
			  (codec->transfer==CODEC_SEND_TRANSFER)
				? "SEND":"RECEIVE");
			*/

			/* set the recovery width and the codec fifo address
			   (read: zoran bitstream reg) at re-start of transfer
			   in case we've changed it for twiddling the zoran
			   during the stall period
			*/
			mga_writeb(codec->mga, CODECCTL+1, codec->rwcfa_shadow);
			
			mgacodec_and_or_codecctl(codec, ctl_mask, ctl);

		} else {
			/*
			printk("mgacodec: %s transfer already started!\n",
			  (codec->transfer==CODEC_SEND_TRANSFER)
				? "SEND":"RECEIVE");
			*/
		}

	} else {	
	/* no transfer has started. start it now.
	   set the transfer direction and reset stalled flag.
	   codec buffer level IRQ will set the stalled flag later.
	   codecrwidth and codecfifoaddr are also set here, but shouldn't be.
	   -inkling */

		/* enable the buffer level IRQ */
		mga_enable_irq(codec->mga, MGA_IRQ_BUFLVL,
				mgacodec_buflevel_irq);

		/* set which direction we're going but i don't see
		   where there is a direction setting in the G200 spec
		   so direction is apparently useless to start/stall/stop
		   -inkling */
		codec->transfer = direction;
		codec->stalled = 0;

		printk("mgacodec: %s transfer starting\n",
		  (codec->transfer==CODEC_SEND_TRANSFER)
			? "SEND":"RECEIVE");

		/* set the recovery width and the codec fifo address 
		   (read: zoran bitstream register) at start of transfer */
		mga_writeb(codec->mga, CODECCTL+1, codec->rwcfa_shadow);

		mgacodec_and_or_codecctl(codec, ctl_mask, ctl);
	}

	return 1;
}

/* this is called by buffer level IRQ to stall the transfer when the hostptr
   limit IRQ has been activated because hardptr has reached hostptr.
   it will try to stall the transfer and set the stalled flag if it did.
   -inkling */
static int mgacodec_stall_transfer(struct mga_codec_s *codec)
{
	u8 ctl_mask = ~(codectransen);
	u8 ctl = 0;

	/* if transfer in progress and transfer not already stalled */
	if (codec->transfer)
	{
		if (!codec->stalled)
		{
			/*
			  printk("mgacodec: stalling %s transfer\n",
			  (codec->transfer==CODEC_SEND_TRANSFER) ? "SEND":"RECEIVE");
			*/

			/* this will clear codectransen and cmdexectrig while
			   keeping all of the other ctl bits the same.
			   i.e. stop the transfer and dont trigger cmd exec.

			   FIXME: if we want to send commands to the zoran during
			   playback, we HAVE to use cmdexectrig during stall period.
			   to do this right will take some IRQ support for cmdcmplpen
			   to tell you when the command is complete and it's ok
			   to enable transfers again.
			   
			   this logic will have to be added because cmdexectrig by
			   itself would stop the transfer and trash current cmd
			   (and data?) fifo.  see G200 spec p 3-69.

			   setting cmdexectrig would clear codectransen and
			   that would NOT reflect in ctl_shadow without some logic.
			   BEST way might be to add a clear for codectransen if the
			   cmdexectrig is set in mgacodec_and_or_ctl to update shadow
			   at least from my interpretation of the G200 spec
			   -inkling */
			
			mgacodec_and_or_codecctl(codec, ctl_mask, ctl);

			udelay(10);

			/* this is the only time codec_stalled status is valid */
			if (!(mga_read(codec->mga, VSTATUS) & codec_stalled))
			{
				printk("mgacodec: unable to stall %s transfer\n",
				       (codec->transfer==CODEC_SEND_TRANSFER)
				       ? "SEND":"RECEIVE");
				return 0;
			}

			codec->stalled = 1;
			/*
			  printk("mgacodec: %s transfer stalled\n",
			  (codec->transfer==CODEC_SEND_TRANSFER)
			  ? "SEND":"RECEIVE");
			*/
		} else {
			/* codec_stalled */
			printk("mgacodec: %s transfer already stalled\n",
			       (codec->transfer==CODEC_SEND_TRANSFER)
			       ? "SEND":"RECEIVE");
		}
	} else {
		/* !codec_transfer */
		printk("mgacodec: no transfer to stall!\n");
	}

	return 1;
}

/* stops the codec interface automatic bitstream transfer
   this stalls any transfer in progress, disables the buffer level IRQ
   and resets the codec. it also clears transfer and stalled flags
   -inkling */
int mgacodec_stop_transfer(struct mga_codec_s *codec)
{
	if (codec->transfer)
	{
		printk("mgacodec: stopping codec transfer\n");
		mgacodec_stall_transfer(codec);

		mga_disable_irq(codec->mga, MGA_IRQ_BUFLVL);
		mgacodec_reset(codec);

		codec->transfer = 0;
		codec->stalled = 0;

		printk("mgacodec: %s transfer stopped\n",
		  (codec->transfer==CODEC_SEND_TRANSFER)
			? "SEND":"RECEIVE");

	} else {
		printk("mgacodec: no transfer to stop!\n");
	}
	return 1;
}

/*
 * calculates how many bytes are left to write new data in the
 * circular buffer
 *
 */ 
static inline int mgacodec_buffer_free(struct mga_codec_s *codec)
{
	int free; /* int32 math ok, it's not greather than +/- 2G */
	free = (mga_read(codec->mga, CODECHARDPTR) << 2) - codec->streamindex;
	if (free < 0) free += CODEC_BUFFER_SIZE;
	free %= CODEC_BUFFER_SIZE; /* not needed? limit it to fit wtf */
/*	printk("mgacodec: free %d hardptr %d, index %d\n", free,
		mga_read(codec->mga, CODECHARDPTR) << 2, codec->streamindex);
*/
	return free;
}

/* program the codec interface to raise an interrupt once a certain 
   byte offset is reached by the bitstream transfer engine
*/
static inline void mgacodec_set_transfer_limit(struct mga_codec_s *codec, u32 hostptr)
{
	/* CODECHOSTPTR must be on an 8 _dword_ boundary (p. 4-248 G100)
	   it is the offset _inside_ the buffer of a _dword_
	*/
	hostptr >>= 2;
	hostptr &= ~0xFFFF0007; /* top 16 bits must be zero G200 spec p 3-72 */
	mga_write(codec->mga, CODECHOSTPTR, hostptr);
}

/* copy a block of data to the host framebuffer for transfer to the
   decoder chip.
*/
static inline void mgacodec_copy_bitstream(struct mga_codec_s *codec, const u8 *bitstream, int size)
{
	int chunk;

	/*
	 * if size if greater than the room left at the end of the buffer,
	 * then we need to split the write operation in two. 1) fill the end of
	 * the buffer, 2) put the remaining bytes at beginning of the ring
	 * (thanks mike@mlb.org for finding it)
	 */
	/* this WOULD help large transfers, but possibly no effect on even
	   2048 byte transfers because the module wrap math SHOULD handle it.
	   if we go with allowing partials, we may get a lot of thrashing
	   or maybe i don't know what i'm talking about :>
	   -inkling */


	chunk = CODEC_BUFFER_SIZE - codec->streamindex;
	if (size > chunk) {
		memcpy_toio((u32)codec->mga->fb + codec->addr +
			    codec->streamindex, bitstream, chunk);
		size -= chunk;
		bitstream += chunk;
		codec->streamindex = 0;
	}

	memcpy_toio((u32)codec->mga->fb + codec->addr + codec->streamindex, bitstream, size);
	codec->streamindex += size;
	codec->streamindex %= CODEC_BUFFER_SIZE;
}


/****************************************************************************
 * PUBLIC INTERFACE                                                         *
 *                                                                          *
 ****************************************************************************/

EXPORT_SYMBOL(mgacodec_init);
EXPORT_SYMBOL(mgacodec_reset);
EXPORT_SYMBOL(mgacodec_and_or_miscctl);
EXPORT_SYMBOL(mgacodec_set_misc2);
EXPORT_SYMBOL(mgacodec_set_ddc2);
EXPORT_SYMBOL(mgacodec_push_readb);
EXPORT_SYMBOL(mgacodec_push_readw);
EXPORT_SYMBOL(mgacodec_push_writeb);
EXPORT_SYMBOL(mgacodec_push_writew);
EXPORT_SYMBOL(mgacodec_get_byte);
EXPORT_SYMBOL(mgacodec_get_word);
EXPORT_SYMBOL(mgacodec_exec_commands);
EXPORT_SYMBOL(mgacodec_stop_transfer);
EXPORT_SYMBOL(mgacodec_send_bitstream);
EXPORT_SYMBOL(mgacodec_buflevel_irq);

/**
 *	mga_codec_init - 	initialize the codec interface
 *	@mga:	mga instance
 *	@addr:	address of a work buffer
 *	@mode:	transfer mode (VMI or I33)
 *	@reg:	destination register for bitstream transfer
 *
 *	This function initializes and configures the Codec Interface
 *	for a particular target chipset. The Interface is set to
 *	IDLE state.
 *
 *	It is the responsability of the programmer to allocate a
 *	work buffer in video memory.
 *
 *	Transfer modes:
 *	%MGACODEC_MODE_I33
 *
 *	%MGACODEC_MODE_VMI_A
 *
 *	%MGACODEC_MODE_VMI_B
 **/
void mgacodec_init(struct mga_dev_s *mga, u32 addr, int mode, int reg)
{
	unique_codec.addr = addr;
	/* p. 3-68 of the MGA HandBook : the Codec Addr must be aligned
	   on a 1Kb boundary
	*/
	unique_codec.addr &= ~0x000003FF;
	unique_codec.ctl_shadow = 0;
	unique_codec.misc_shadow = 0;
	unique_codec.cmdindex = 0;
	unique_codec.dataindex = 0;
	unique_codec.streamindex = 0;
	unique_codec.transfer = 0;
	init_waitqueue_head(&unique_codec.outq);
	init_waitqueue_head(&unique_codec.inq);
	unique_codec.mga = mga;
	unique_codec.mode = mode;
	unique_codec.dstreg = reg;

	mga->codec = &unique_codec;	/* bind to static instance */

	mgacodec_reset(&unique_codec);
}

/**
 *	mgacodec_reset -	reset the Codec Interface
 *	@codec:	codec instance
 *	
 *	This function resets the Codec Interface to the IDLE state,
 *	disabling transfers.
 *
 *	The order for the 3 CODECCTL bytes should probably be (need to
 *	verify):
 *
 *	1. clear codec general control
 *
 *	2. set recovery width and codecfifoaddr (only at (re)start
 *	transfer?)
 *
 *	3. set codec general control
 *
 *	4. set codecaddr
 *
 *	-inkling
 **/
void mgacodec_reset(struct mga_codec_s *codec)
{
	printk("mgacodec: starting reset\n");
	codec->transfer = 0;	/* transfer off */
	codec->stalled = 0;	/* stalled off */

	/* clear command completion and buffer level interrupt pending.
	   we're not using command completion interrupts yet, but we
	   will want to eventually use them. does this even belong here?
	   -inkling */
	mga_write(codec->mga, VICLEAR, blvlpen | cmdcmplpen);

	/* 1. disable codec.  see G200 spec, p 4-89, section 4.9.5 */
	mga_writeb(codec->mga, CODECCTL, 0);

	/* see G200 spec, p 4-89ff
	   2. set codec recovery width BEFORE setting codecen
	   and set codecfifoaddr to 4 (zoran bitstream register)
	   FIXME: do this here, or only at codec_start_transfer?
	   -inkling */
	codec->rwcfa_shadow = (codec->rwidth<<4) | (codec->dstreg & 15);
	mga_writeb(codec->mga, CODECCTL+1, codec->rwcfa_shadow);

	/* I repeat it twice per p. 4-110 of the G400 HB */
	/* could find no corresponding reference in G200 spec
	   but it shouldn't hurt anything on the G200, so wtf.
	   maybe it's some kind of byte/word transfer issue on AGP. who knows?
	   -inkling */
	/* set codecen to enable the codec
	   these were incorrectly using miscctl before, since 1.2 cvs.
	   look at 1.1 and 1.2 cvs of this file to see what i mean
	   -inkling */
	mgacodec_and_or_codecctl(codec, codecen, codecen);
	mgacodec_and_or_codecctl(codec, codecen, codecen);

	/* see G200 spec p 3-68: bit0=1=256k buffer, bit0=0=128k buffer */
	mga_write(codec->mga, CODECADDR,
		  (codec->addr & ~0xF00003FE ) |
		  (CODEC_BUFFER_SIZE == 0x40000 ? 1 : 0) );

	printk("mgacodec: reset codec done\n");
}

/**
 *	mgacodec_and_or_miscctl -	twiddle MISCCTL
 *	@codec:	codec instance
 *	@and:	AND operand
 *	@or:	OR operand
 *
 * 	This function twiddles the MISCCTL part of the codec interface
 *	to reset the actual CODEC chip connected to the MISC lines of
 *	the interface.
 *
 * 	QUESTION: should this be called AFTER recovery width setting?
 *	G200 docs say to set recovery width BEFORE enabling codec.
 *	this code enables the codec.
 *	-inkling
 **/
void mgacodec_and_or_miscctl(struct mga_codec_s *codec, u8 and, u8 or)
{
	/* see G200 spec p 4-90ff for compression information
	   see G200 spec p 4-93ff for decompression information
	   we must be in IDLE (or COMP/DECOMP) for this to work

	   decompression:
	   p. 4-92 states that MISC<2> is connected to EOI (?) when the
	   CODEC Interface is enabled.
	   So it seems that we can bang on these 2 bits too...
	*/

	/* we MUST either verify we are in IDLE (or COMP/DECOMP), or force
	   the right state before doing this. see G200 spec p 4-89ff
	   we should send a 0 to the low byte of codecctl (to stop the codec;
	   optional but GOOD IDEA) then send a 1 to the low byte of codecctl
	   (to enable the codec) and then send the misctl bits to the high
	   byte of codecctl. the procedure is detailed in G200 spec p 4-94ff
	   -inkling */

	/* this follows the G200 spec p 4-89ff
	   disable (spec says optional, but should force IDLE state)
	   then enable the codec, then program the MISCCTL byte.
	   also, update ctl shadow with new value since we changed CODECCTL+0
	   and update the misc shadow too.
	   -inkling */
	mga_writeb(codec->mga, CODECCTL, 0);
	mga_writeb(codec->mga, CODECCTL, 1);
	codec->ctl_shadow = 1;

	codec->misc_shadow &= and;
	codec->misc_shadow |= or;
	mga_writeb(codec->mga, CODECCTL+3, codec->misc_shadow);
}

/**
 *	mgacodec_set_misc2 -	program the MISC<2> line
 *	@codec:	codec instance
 *	@param:	0 to disable the output driver, 1 to enable the output driver
 *
 * 	Set the mode of the MISC<2> line (input or output).
 *
 *	See:	p. 3-325 (361) of the G200 spec
 **/
int mgacodec_set_misc2(struct mga_codec_s *codec, u8 param)
{
	mgacodec_and_or_codecctl(codec, 0xFF, codecen | codec->mode);
	mgacodec_and_or_miscctl(codec, ~param, 0);
	mgacodec_and_or_codecctl(codec, ~codecen, 0);
	mga_indirect_and_or(codec->mga, XGENIOCTRL, ~0x40, 0x00);
	if (mga_indirect_readb(codec->mga, XGENIOCTRL) & 0x40) {
		printk("mgacodec: error setting misc2 line\n");
		return 0;
	} else
		return 1;
}

/**
 *	mgacodec_set_ddc2 -	program the DDC<2> line
 *	@codec:	codec instance
 *	@param:	0 to disable the output driver, 1 to enable the output driver
 *
 * 	Set the mode of the DDC<2> line (input or output).
 *
 * 	Rationale: p. 5-69 (389) and p. 5-92 (392) of the G_1_00 spec
 * 	The DDC<2> line is connected to the HRDY line for a DVD addon board 
 **/
int mgacodec_set_ddc2(struct mga_codec_s *codec, u8 param)
{
	mgacodec_and_or_codecctl(codec, 0xFF, codecen | codec->mode);
	mgacodec_and_or_miscctl(codec, ~param, 0);
	mgacodec_and_or_codecctl(codec, ~codecen, 0);
	mga_indirect_and_or(codec->mga, XGENIOCTRL, ~0x04, 0x00);
	writeb(XGENIOCTRL, codec->mga->ctrl + PALWTADD);
	writeb(readb(codec->mga->ctrl + X_DATAREG) & ~0x04, codec->mga->ctrl + X_DATAREG);
	if (mga_indirect_readb(codec->mga, XGENIOCTRL) & 0x04) {
		printk("mgacodec: error setting ddc2 line\n");
		return 0;
	} else
		return 1;
}

/**
 * 	mgacodec_push_readb -	add a read instruction (8 bits)
 *	@codec:	codec instance
 *	@reg:	register to read
 *
 *	Add a read instruction to the command buffer.
 *	To actually read a register, you must call
 *	mgacodec_exec_commands() and mgacodec_get_word() after this.
 **/
void mgacodec_push_readb(struct mga_codec_s *codec, u8 reg)
{
#ifdef RRG_TRICK
	reg = RRG_TRICK(reg);
#endif
	codec->cmdbuffer[codec->cmdindex++] = (1<<15) | (reg<<10) | 1;
	codec->cmdbuffer[codec->cmdindex++] = 0;
}

/**
 * 	mgacodec_push_readw -	add a read instruction (16 bits)
 *	@codec:	codec instance
 *	@reg:	register to read
 *
 *	Add a read instruction to the command buffer.
 *	To actually read a register, you must call
 *	mgacodec_exec_commands() and mgacodec_get_word() after this.
 **/
void mgacodec_push_readw(struct mga_codec_s *codec, u8 reg)
{
#ifdef RRG_TRICK
	reg = RRG_TRICK(reg);
#endif
	codec->cmdbuffer[codec->cmdindex++] = (1<<15) | (reg<<10) | 2;
	codec->cmdbuffer[codec->cmdindex++] = 0;
}

/**
 * 	mgacodec_push_writeb -	add a write instruction (8 bits)
 *	@codec:	codec instance
 *	@reg:	register to write
 *	@data:	byte to write
 *
 *	Add a write instruction to the command buffer.
 *	To actually write the data, you must call
 *	mgacodec_exec_commands() after this.
 *
 *	See also:	mgacodec_push_readb(),
 *	mgacodec_push_writeb(), mgacodec_exec_commands()
 **/
void mgacodec_push_writeb(struct mga_codec_s *codec, u8 reg, u8 data)
{
#ifdef RRG_TRICK
	reg = RRG_TRICK(reg);
#endif
	codec->cmdbuffer[codec->cmdindex++] = 1;
	codec->cmdbuffer[codec->cmdindex++] = 0;

	codec->cmdbuffer[codec->cmdindex++] = (reg << 8) | data;
	codec->cmdbuffer[codec->cmdindex++] = 0;
}

/**
 * 	mgacodec_push_writew -	add a write instruction (16 bits)
 *	@codec:	codec instance
 *	@reg:	register to write
 *	@data:	word to write
 *
 *	Add a write instruction to the command buffer.
 *	To actually write the data, you must call
 *	mgacodec_exec_commands() after this.
 *
 *	See also:	mgacodec_push_readw(),
 *	mgacodec_push_writew(), mgacodec_exec_commands()
 *
 *	Notes: this function is tuned to transfer words to a
 *	big-endian chip (like the ZR367x0 chipset).
 *
 *	Fixme: Only i386 has been tested so far.
 **/
void mgacodec_push_writew(struct mga_codec_s *codec, u8 reg, u16 data)
{
#ifdef RRG_TRICK
	reg = RRG_TRICK(reg);
#endif
	codec->cmdbuffer[codec->cmdindex++] = 2;
	codec->cmdbuffer[codec->cmdindex++] = 0;

	codec->cmdbuffer[codec->cmdindex++] = (CODEC_WRITE(reg, (data & 0x00FF))<<16)
		| CODEC_WRITE(reg, ((data>>8) & 0xFF));
	codec->cmdbuffer[codec->cmdindex++] = 0;
}

/**
 * 	mgacodec_push_pause -	add a pause in the command stream
 *	@codec:	codec instance
 *
 *	This function sets the pause bit in the last 'pushed
 * 	command stream, as described in the Matrox spec. Once the
 *	Codec Interface see the 'pause' bit it will automtically
 *	enable data transfer (read or write).
 *
 *	See also:	mgacodec_push_readb/w(),
 *	mgacodec_push_writeb/w(), and mgacodec_exec_commands()
 **/
void mgacodec_push_pause(struct mga_codec_s *codec)
{
#ifdef RRG_TRICK
	reg = RRG_TRICK(reg);
#endif
	if (codec->cmdindex < 4) return;

	codec->cmdbuffer[codec->cmdindex - 4] |= (1<<8);
}

/**
 *	mgacodec_get_byte -	get the result of the last commands (8 bits)
 *	@codec:	codec instance
 *
 *	This function reads one byte in the data buffer returned by
 *	previous read instructions. Successive calls dequeue the bytes
 *	received in the data buffer.
 *
 *	Return value:	a byte
 *
 *	See also:	mgacodec_push_readb(),
 *	mgacodec_push_writeb(), mgacodec_exec_commands()
 **/
u8 mgacodec_get_byte(struct mga_codec_s *codec)
{
	u8 data;

	data = codec->databuffer[codec->dataindex] & 0x00FF;
	codec->dataindex += 4;

	return data;
}

/**
 *	mgacodec_get_word -	get the result of the last commands (16 bits)
 *	@codec:	codec instance
 *
 *	This function reads one word in the data buffer returned by
 *	previous read instructions. Successive calls dequeue the word
 *	received in the data buffer.
 *
 *	Return value:	a word
 *
 *	See also:	mgacodec_push_readw(),
 *	mgacodec_push_writew(), mgacodec_exec_commands()
 *
 *	Notes: this function is tuned to transfer words to a
 *	big-endian chip (like the ZR367x0 chipset).
 *
 *	Fixme: Only i386 has been tested so far.
 **/
u16 mgacodec_get_word(struct mga_codec_s *codec)
{
	u16 data;

	data = __swab16(codec->databuffer[codec->dataindex]);
	codec->dataindex += 4;

	return data;
}

/**
 *	mgacodec_exec_commands - 	executes queued commands
 *	@codec:	codec instance
 *
 * 	This function executes the commands queued in the command
 *	buffer by previous mgacodec_push_read() (or _write()) calls.
 *
 *	It takes care of suspending any bitstream transfer that may
 *	take place at the time it is called, without loosing datas (it
 *	first asks the codec interface to stall transfer and waits for
 *	it to complete). This function also automatically restarts an
 *	interrupted transfer after commands have been executed.
 *
 */
int mgacodec_exec_commands(struct mga_codec_s *codec)
{
	/*
	 * FIXME: horrible hack to make mgadvd work !!!!
	 * without this, the commands hanged...
	 */
#ifdef MGADVD_HACK
	mgacodec_push_readw(codec, 0x7);		/* ZR_DEVICE_ID */
#endif
	codec->cmdbuffer[codec->cmdindex-2] |= (1<<9);	/* activate the stop bit */

	/* copy command buffer into video memory */
	memcpy_toio((u32)codec->mga->fb + codec->addr + CODEC_BUFFER_SIZE,
			 codec->cmdbuffer, 256);

	/* check that bitstream transfer is stalled, so that we can safely
	   trigger command execution
	   (the rationale for this process comes from the G100 specs p. 5-90)
	*/
	/* mgacodec_stall_transfer(codec); // <--- DO NOT DO THIS */
	/* codec is stopped by cmdexectrig and fifo trashed. doing a stall
	   here only results in an endless loop if a transfer is not in
	   progress, i.e. initial state before any bitstream is fed.
	   see G200 spec p.3-198 and p. 3-69
	   wanna see klogd reboot your machine? leave it in with the
	   printk's that i've added to codec_stall.
	   -inkling */
	

//	mgacodec_and_or_codecctl(codec, codecen | cmdexectrig | codecmode,
//				 codecen | cmdexectrig | codec->mode);

	mgacodec_and_or_codecctl(codec, 0, codecen | cmdexectrig);
	
	codec->dataindex = 0;
	codec->cmdindex = 0;

	/* give some time to the Codec Interface to execute commands */
	udelay(10);
	memcpy_fromio(codec->databuffer, (u32)codec->mga->fb + codec->addr + CODEC_BUFFER_SIZE + 0x200, 256);

	if (codec->transfer)
//		mgacodec_and_or_codecctl(codec, codecen | codectransen | codecmode,
//					 codecen | codectransen | codec->mode);

		mgacodec_and_or_codecctl(codec, 0, codecen | codectransen);

	return 1;
}

/**
 *	mgacodec_send_bitstream -	initiates bitstream transfer
 *	@codec:		codec instance
 *	@bitstream:	block to send
 *	@size:		size of the block
 *
 *	This function initiates bitstream transfer to the decoder
 *	chip. It writes the bitstream to the Codec Interface buffer
 *	(using the interrupt handler). More specifically, what we do
 *	here is :
 *
 *	    - check if there is room left to write or _wait_ until
 *	      there is;
 *
 *	    - write a block of data;
 *
 *	    - tell the interrupt handler to stall the transfer if the
 *	      codec interface runs past the last written block.
 */
int mgacodec_send_bitstream(struct mga_codec_s *codec, const u8 *bitstream, unsigned int size)
{
	u32 buf_free;
	u32 buf_limit;

	/*
	  Codec Bitstream Buffer
	  [..........................................................]
	            ^                ^         ^
		    hardptr          hostptr   streamindex

		  o  hardptr is driven by the hardware: it's a pointer
		     to the qword currently sent to the decoder
		  o  hostptr is handled by our driver: it triggers an
		     IRQ to stall the codec when hardptr reaches it
		  o  streamindex is handled by our driver: it marks
		     the point in the codec buffer where we're
		     currently writing
	*/

	/* if the transfer has not yet started, prefill the buffer
	    until it is nearly full before we start the transfer
	    -inkling */
	if (!codec->transfer)
	{
		// printk("mgacodec: prefill index %d\n", codec->streamindex);

		mgacodec_copy_bitstream(codec, bitstream, size);
		if (codec->streamindex <= CODEC_BUFFER_SIZE - 16384)
		{
			/* buffer is not full yet, keep filling */
			return 1;
		} else {
		/* buffer is full enough. set hostptr IRQ to stall
		   transfer when this pointed reached, then start transfer.
		   -inkling */
			buf_limit = (codec->streamindex - 16384) % CODEC_BUFFER_SIZE;
			mgacodec_set_transfer_limit(codec, buf_limit);
			mgacodec_start_transfer(codec, CODEC_SEND_TRANSFER);
			return 1; /* get more data and possibly have to wait */
		}

	}
	/* the transfer has started, check if we can add more data.
	   if we can't add more data, we will have to wait for the
	   codec buffer level IRQ to indicate that some space is free.
	   -inkling */

	buf_free = mgacodec_buffer_free(codec);

	if ( buf_free < size )
	{
	/* we have to wait for some space to be freed up by waiting
	   until the codec reaches buffer level IRQ limit and stalls.
	   set the interrupt to happen, then wait for it.
	   -inkling */

		buf_limit = (codec->streamindex - 8192) % CODEC_BUFFER_SIZE;
		mgacodec_set_transfer_limit(codec, buf_limit);
		interruptible_sleep_on(&codec->outq);

		/* abort if we received any signal during this wait */
		if (signal_pending(current))
		{
			printk("mgacodec: wait interrupted by signal\n");
			return -ERESTARTSYS;
		}

	}
        
	/* write the data to the buffer and adjust the buffer level IRQ
	   to stall the codec once it is reached. this keeps any stalls
	   in the feed stream from corrupting the playback.

	   this provides the very much desired benefit of allowing the
	   codec to gracefully handle hardware induced stalls in the feed
	   stream (dvd sector read pause at layer change or task switching.)

	   -inkling */

	mgacodec_copy_bitstream(codec, bitstream, size);
	buf_limit = (codec->streamindex - 4096) % CODEC_BUFFER_SIZE;
	mgacodec_set_transfer_limit(codec, buf_limit);
	mgacodec_start_transfer(codec, CODEC_SEND_TRANSFER);
	return 1;	
}

/**
 *	mgacodec_recv_bitstream -	initiates bitstream receiving
 *	@codec:		codec instance
 *
 *	FIXME:	write me !
 */
int mgacodec_recv_bitstream(struct mga_codec_s *codec)
{
	return 0;
}

/**
 *	mgacodec_buflevel_irq -	handle BLVL interrupts
 *	@mga:	mga instance
 *	@regs:	general registers passed to an interrupt handler
 *
 *	This routine handles BLVL interrupts. It is called each time
 *	the Codec Interface transfer engine has reached a programmed
 *	limit. This routine just stalls the bitstream transfer to
 *	avoid data loss and bitstream corruption and wakes up queue
 *	tasks. For example, if the Codec Interface works in
 *	SEND_TRANSFER mode (decompression), then this routine will
 *	signal the 'outq' wait queue.
 *
 *	See also:	mga_enable_irq() to install the routine
 *
 *	Note:	this routine is installed by default when the
 *	mgacodec_send_bitstream() function is called repeatedly. This
 *	will change to accommodate the recv_bitstream variant.
 */
void mgacodec_buflevel_irq(struct mga_dev_s *mga, struct pt_regs *regs)
{
	struct mga_codec_s *codec = mga->codec;

	mgacodec_stall_transfer(codec);
	if (codec->transfer == CODEC_SEND_TRANSFER)
		wake_up_interruptible(&codec->outq);
	if (codec->transfer == CODEC_RECV_TRANSFER)
		wake_up_interruptible(&codec->inq);
}

/**
 *	mgacodec_poll -	handle poll syscall
 *	@codec:	codec_instance
 *	@file:	see &file_operations
 *	@wait:	see &file_operations
 *	@read_threshold:	minimum number of bytes to consider
 *				that read will succeed
 *	@write_threshold:	minimum number of bytes to consider
 *				that write will succeed
 *
 *	mgacodec_poll() is a helper function for driver implementing
 *	the 'poll' file operation. The threshold parameters help
 *	ensure that a subsequent 'write' system call won't block
 *	because there isn't enough space left.
 *
 **/
unsigned int mgacodec_poll(struct mga_codec_s *codec, struct file *file, poll_table *wait,
			   u32 read_threshold, u32 write_threshold)
{
	unsigned int mask;

	poll_wait(file, &codec->inq, wait);
	poll_wait(file, &codec->outq, wait);
	mask = 0;
	if (mgacodec_buffer_free(codec) > write_threshold)
                mask |= POLLOUT | POLLWRNORM;

	/* FIXME: add more stuff here when compressing video */

	return mask;
}

