/*
 * This module provides an interface to the core video capabilities
 * of the G100/G200/G400 chipset.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/videodev.h>
#include "mgacore.h"

MODULE_AUTHOR("David Barth <dbarth@besancon.net>");
MODULE_DESCRIPTION("Core Matrox module");

EXPORT_SYMBOL(mga_get_card);
EXPORT_SYMBOL(mga_release_card);
EXPORT_SYMBOL(mga_alloc_buffer);
EXPORT_SYMBOL(mga_free_buffer);
EXPORT_SYMBOL(mga_enable_irq);
EXPORT_SYMBOL(mga_disable_irq);
EXPORT_SYMBOL(mga_request_irq);
EXPORT_SYMBOL(mga_release_irq);

#define MAX_CARDS	4
static struct mga_dev_s card[MAX_CARDS];

/* * PRIVATE FUNCTIOM *
 *	mga_interrupt_handler -	main interrupt handler
 *	@mga:		mga instance
 *	@dev_id:	
 *	@regs:
 *
 *	This function handles interrupt requests. It will call
 *	registered sub-handlers for each specific interrupt.
 *	
 *	See also:
 *	mga_request_irq()
 *
 *	mga_release_irq()
 *
 *	Notes:
 *	mga_interrupt_hw() is responsible for clearing the
 *	interrupt. Sub-handlers should not worry about it.
 */
static void mga_interrupt_handler(int irq, void *dev_id,
				  struct pt_regs *regs)
{
	u32 status, vstatus;
	struct mga_dev_s *mga = dev_id;

	if (mga == NULL) {
		printk("mgacore: mga interrupt handler says *mga is NULL!\n");
		return;
	}

	spin_lock(mga->lock);
        /* read status register */
	vstatus = readl(mga->ctrl + VSTATUS);
	status  = readl(mga->ctrl + STATUS);

	if (vstatus & MGA_IRQ_VINVSYNC) {
		if (mga->irq_handler[MGA_IRQ_VINVSYNC])
			(*mga->irq_handler[MGA_IRQ_VINVSYNC])(mga, regs);
		mga_write(mga, VICLEAR, MGA_IRQ_VINVSYNC);
	}

	if (vstatus & MGA_IRQ_BUFLVL) {
		/* disable all mga video interrupts - should be already 
		mga_write(mga, VIEN, 0); */

		if (mga->irq_handler[MGA_IRQ_BUFLVL])
			(*mga->irq_handler[MGA_IRQ_BUFLVL])(mga, regs);

		/* clear any pending buffer level interrupts */
		mga_write(mga, VICLEAR, MGA_IRQ_BUFLVL);

		/* re-enable buffer level interrupts */
		mga_write(mga, VIEN, MGA_IRQ_BUFLVL);
	}

	if (vstatus & MGA_IRQ_DCMPEOI) {
		if (mga->irq_handler[MGA_IRQ_DCMPEOI])
			(*mga->irq_handler[MGA_IRQ_DCMPEOI])(mga, regs);
		mga_write(mga, VICLEAR, MGA_IRQ_DCMPEOI);
	}


	if (status & MGA_IRQ_EXTERNAL) {
		if (mga->irq_handler[MGA_IRQ_EXTERNAL])
			(*mga->irq_handler[MGA_IRQ_EXTERNAL])(mga, regs);
		/* FIXME: clear interrupt */
		/* no fix needed. reading status clears the interrupt
		   -inkling */

	}

	spin_unlock(mga->lock);
}

/**
 *	mga_request_irq - install an interrupt handler for PCI IRQ
 *	@mga:		mga instance
 *	@handler:	interrupt handler function
 *
 *	This function requests the PCI IRQ used by the card from
 *	the kernel and installs an interrupt handler to manage
 *	all of the various IRQ events that may occur.
 *
 *	All card interrupts are disabled BEFORE the request so that
 *	no spurious interrupts will occur before any of the
 *	sub-handlers have been initialized.
 *	
 *	If another handler is already installed, this function will
 *	NOT overwrite it. The other handler must be released before.
 *
 *	IRQs:
 *	%MGA_IRQ_VINVSYNC
 *
 *	%MGA_IRQ_BUFLVL
 *
 *	%MGA_IRQ_DCMPEOI
 *
 *	%MGA_IRQ_EXTERNAL
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	See also:
 *			mga_release_irq()
 *			mga_enable_irq()
 *			mga_disable_irq()
 */
int mga_request_irq(struct mga_dev_s *mga)
{
	unsigned long flags;
	u32 vien_chk;

	spin_lock_irqsave(mga->lock, flags);

	if (!mga->irq_inuse)
	{
		/* see G200 spec, Table 4-8 on p 4-115
		   clear and disable the 4 MGA video IRQ sources. this may
		   be overkill but better too much caution than not enough.
		   -inkling */
		mga_write(mga, VICLEAR, 0);
		mga_write(mga, VIEN, 0);

		/* all MGA Video interrupts should be disabled now. we
		   need to check them and if they are not disabled,
		   return an error for total abort of process
		   -inkling */
		
		vien_chk = mga_read(mga, VIEN) & 0xF; /* mask reserved bits */

		if (vien_chk)
		{
			printk("mgacore: PCI IRQ request could not clear VIEN (0x0%1X)\n",
				vien_chk);
			return -ENODEV;
		}

		/* FIXME: add clears and disables for whichever of the other 7
			   MGA graphic engine IRQ sources we want to use */

		if (request_irq(mga->dev->irq, mga_interrupt_handler,
				SA_INTERRUPT | SA_SHIRQ, "mgacore", mga) < 0)
		{
			  printk("mgacore: denied PCI IRQ %d\n",  mga->dev->irq);
			  return -ENODEV;
		}
		mga->irq_inuse = 1; /* set the IRQ In Use counter to 1 */
		printk("mgacore: PCI IRQ %d handler registered\n",
			mga->dev->irq);
	} else {
		printk("mgacore: PCI IRQ %d handler is already registered\n",
			mga->dev->irq);
	}
	spin_unlock_irqrestore(mga->lock, flags);

	return 0;
}

/**
 *	mga_release_irq - release a PCI IRQ interrupt handler
 *	@mga:	mga instance
 *
 *	This function releases the PCI IRQ handler used by the card
 *	after disabling all enabled interrupt sources on the card.
 *	Any associated sub-handlers are indicated upon release.
 *
 *	See also:
 *			mga_request_irq()
 *			mga_enable_irq()
 *			mga_disable_irq()
 *
 *	FIXME: all the sub-handler pointers should also be set to NULL.
 * 		it's not really a crucial step, but more a neatness thing
 *		-inkling
 */
int mga_release_irq(struct mga_dev_s *mga)
{
	unsigned long flags;
	u32 vien_chk;

	spin_lock_irqsave(mga->lock, flags);

	if (mga->irq_inuse)
	{
		printk("mgacore: releasing PCI IRQ %d handler\n",
			mga->dev->irq);

		if (mga->irq_inuse > 1)
		{
			printk("mgacore: %d IRQ sub-handlers will be affected!\n",
				mga->irq_inuse - 1);
		}
		/* see G200 spec, Table 4-8 on p 4-115
		   clear and disable the 4 MGA video IRQ sources. this may
		   be overkill but better too much caution than not enough.
		   -inkling */
		mga_write(mga, VICLEAR, 0);
		mga_write(mga, VIEN, 0);

		/* all MGA Video interrupts should be disabled now. now we
		   need to check them and if they are not disabled,
		   return an error
		   -inkling */
		
		vien_chk = mga_read(mga, VIEN) & 0xF; /* mask reserved bits */

		if (vien_chk)
		{
			printk("mgacore: PCI IRQ %d release could not clear VIEN (0x0%1X)\n",
				mga->dev->irq, vien_chk);
			return -ENODEV;
		}

		/* FIXME: add clears and disables for whichever of the other 7
			   MGA graphic engine IRQ sources we want to use */

		free_irq(mga->dev->irq, mga);
		printk("mgacore: PCI IRQ %d released\n", mga->dev->irq);
	} else {
		printk("mgacore: no registered PCI IRQ handler to release!\n");
	}			

	spin_unlock_irqrestore(mga->lock, flags);

	return 0;
}

/**
 *	mga_enable_irq - enable an interrupt source to use a sub-handler
 *	@mga:		mga instance
 *	@irq:		IRQ to enable
 *
 *	This function enables the hardware to generate interrupts from
 *	the specified source.
 *
 *	This will NOT allow you to replace a sub-handler.  You must disable
 *	it first if you want to replace it with a new sub-handler.
 *
 *	FIXME: We may want to modify this to allow sub-handler replacement.
 *	-inkling	
 *
 *	IRQs:
 *	%MGA_IRQ_VINVSYNC
 *
 *	%MGA_IRQ_BUFLVL
 *
 *	%MGA_IRQ_DCMPEOI
 *
 *	%MGA_IRQ_EXTERNAL
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	See also:
 *			mga_disable_irq()
 *			mga_request_irq()
 *			mga_release_irq()
 */
int mga_enable_irq(struct mga_dev_s *mga, int irq, mga_interrupt_handler_t sub_handler)
{
	unsigned long flags;
	u32 vien, vien_chk;

	spin_lock_irqsave(mga->lock, flags);

	/* first we disable and clear the interrupt we're going to use
	   so it doesn't crash when we replace it, even though interrupts
	   are supposed to be off right now. better to be cautious
	   -inkling */

	vien = mga_read(mga, VIEN) & ~irq;
	mga_write(mga, VIEN, vien);
	mga_write(mga, VICLEAR, irq);
	vien_chk = mga_read(mga, VIEN);

	/* FIXME: this is properly cautious, but the printk's need to be
		  commented once it is verified to be working.
		  -inkling */
	if (vien_chk && irq)
	{ 
		printk("mgacore: enable VIRQ failed to clear VIRQ 0x%1X\n",
		irq);
		return -ENODEV;
	} else {
		printk("mgacore: enable VIRQ has cleared VIRQ 0x%1X 0x%1X\n",
		irq, vien);
	}
 
	/* next we assign the sub-handler pointer and increment in_use count.
	   no change and no error (except a klog msg) if a sub-handler
	   is already registered and try to re-enable the MGA Video IRQ.
	   -inkling */

	if (!mga->irq_handler[irq])
	{
		mga->irq_handler[irq] = sub_handler;
		mga->irq_inuse++;
	} else {
		printk("mgacore: sub-handler for VIRQ 0x%1X already exists!\n",
			irq);
	}

	/* last step is to clear again any pending video irq (cautious)
	   and enable the interrupt to use the handler
	   -inkling */

	vien = mga_read(mga, VIEN) | irq;
	mga_write(mga, VICLEAR, irq);
	mga_write(mga, VIEN, vien);

	/* check to see if it actually did enable the interrupt.
	   return failure code if it did NOT (re)enable the interrupt.
	   -inkling */

	vien_chk = mga_read(mga, VIEN);

	/* FIXME: this is properly cautious, but the printk's need to be
		  commented once it is verified to be working.
		  -inkling */
	if (!(vien_chk & irq))
	{
		printk("mgacore: VIRQ %d could not be enabled!\n", irq);
		return -ENODEV;
	} else {	
		printk("mgacore: enabled VIRQ %d\n", irq);
	}

	spin_unlock_irqrestore(mga->lock, flags);
	return 0;
}


/**
 *	mga_disable_irq - disable an interrupt source
 *	@mga:	mga instance
 *	@irq:	IRQ to disable
 *
 *	This function disables the generation of hardware interrupts from
 *	the specified source.
 *
 *	The associated interrupt handler IS uninstalled by this
 *	function.
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	See also:
 *			mga_enable_irq()
 *			mga_request_irq()
 *			mga_release_irq()
 */
int mga_disable_irq(struct mga_dev_s *mga, int irq)
{
	unsigned long flags;
	u32 vien;

	spin_lock_irqsave(mga->lock, flags);

	/* G200 spec say clear pending interrupts before enable. it
	   does NOT say if it needs to be done before disabling. it would
	   be cautious to clear any pending interrupt before disabling.
	   -inkling */

	mga_write(mga, VICLEAR, irq);
	vien = mga_read(mga, VIEN) & ~irq;
	mga_write(mga, VIEN, vien);
	vien = mga_read(mga, VIEN);
	if (vien & irq)
	{
		printk("mgacore: disable VIRQ is unable to clear VIRQ %d\n",
			 irq);
		return -ENODEV;
	}		

	if (mga->irq_handler[irq])
	{
		mga->irq_handler[irq] = NULL;
		mga->irq_inuse--;
	} else {
		printk("mgacore: no sub-handler for VIRQ 0x%1X to remove!\n",
			irq);
	}

	spin_unlock_irqrestore(mga->lock, flags);
	return 0;
}




/* * PRIVATE FUNCTION
 *      mga_test_lines - print I/O lines status
 *      @mga:     mga instance
 *      @details: if non null print more details
 *
 *      Debugging function printing the state of the general I/O lines.
 */
static void mga_test_lines(struct mga_dev_s *mga, int details)
{
	u8 ddcoe, miscoe;
	u8 ddcdata, miscdata;
	u8 xgenioctrl, xgeniodata;

	xgenioctrl = mga_indirect_readb(mga, XGENIOCTRL);
	ddcoe  = xgenioctrl & 0x0F;
	miscoe = (xgenioctrl & 0x70)>>4;
	xgeniodata = mga_indirect_readb(mga, XGENIODATA);
	ddcdata  = xgeniodata & 0x0F;
	miscdata = (xgeniodata & 0x70)>>4;

	if (details) {
		printk("mgacore: I/O lines summary\n");

		printk("mgacore: DDC<0> %s %d\n", (ddcoe & 1) ? "->" : "<-", ddcdata & 1);
		printk("mgacore: DDC<1> %s %d\n", (ddcoe & 2) ? "->" : "<-", (ddcdata & 2)>>1);
		printk("mgacore: DDC<2> %s %d\n", (ddcoe & 4) ? "->" : "<-", (ddcdata & 4)>>2);
		printk("mgacore: DDC<3> %s %d\n", (ddcoe & 8) ? "->" : "<-", (ddcdata & 8)>>3);

		printk("mgacore: MISC<0> %s %d\n", (miscoe & 1) ? "->" : "<-", miscdata & 1);
		printk("mgacore: MISC<1> %s %d\n", (miscoe & 2) ? "->" : "<-", (miscdata & 2)>>1);
		printk("mgacore: MISC<2> %s %d\n", (miscoe & 4) ? "->" : "<-", (miscdata & 4)>>2);
	} else {
		printk("mgacore: DDC: %d%d%d%d MISC: %d%d%d\n",
		       ddcdata & 1, (ddcdata & 2)>>1, (ddcdata & 4)>>2, (ddcdata & 8)>>3,
		       miscdata & 1, (miscdata & 2)>>1, (miscdata & 4)>>2);
	}
}

/**
 *	mga_alloc_buffer - 	allocate a buffer
 *	@mga:	mga instance
 *	@size:	size of the buffer to allocate
 *
 *	This is a really simple memory manager for the framebuffer.
 *	It justs decrements a pointer from the end of the framebuffer
 *	and DOES NOT really manage an array of allocated blocks.
 *	
 *	An allocated block must be deallocated in the SAME ORDER (with
 *	respect to other modules requesting memory) for this to work.
 *
 *	Return value:	returns a pointer to the allocated buffer
 *
 *	See also:	mga_free_buffer()
 *
 *	Notes:	The memory address is expressed as a value *inside*
 *	the framebuffer address space. For example a buffer at address
 *	0x0 represents the beginning of the video framebuffer.
 */
void * mga_alloc_buffer(struct mga_dev_s *mga, unsigned long size)
{
	if (mga->malloc_limit < size)
		return NULL;

	mga->malloc_limit -= size;
	printk("mgacore: alloc %ld (%#lx)\n", size, size);
	return (void *)mga->malloc_limit;
}

/**
 *	mga_free_buffer - 	deallocate a buffer
 *	@mga:	mga instance
 *	@size:	size of the block to allocate
 *
 *	Increments the allocation counter, to simulate memory
 *	deallocation.
 *
 *	An allocated block must be deallocated in the SAME ORDER (with
 *	respect to other modules requesting memory) for this to work.
 *
 *	See also:	mga_alloc_buffer()
 */
void mga_free_buffer(struct mga_dev_s *mga, unsigned long size)
{
	mga->malloc_limit += size;
	printk("mgacore: free %ld (%#lx)\n", size, size);
}

/* * PRIVATE FUNCTION *
 *	mga_guess_memory_size -	detect frambuffer size
 *	@mga:   mga device structure
 *
 *	This function queries the graphic card to detect the size of
 *	the framebuffer.
 *
 *	Return value:
 *	Memory size
 *
 *	Notes:
 * 	This memory sizer is from the Matrox framebuffer driver
 *	developed by Petr Vandrovec <vandrove@vc.cvut.cz>,
 *	et. al. Thanks for porting the code Steve. 
 */
static unsigned int mga_guess_memory_size(struct mga_dev_s *mga)
{
	unsigned int offs;
        unsigned int offs2;
        unsigned char store;
        unsigned char bytes[32];
        unsigned char* tmp;
	unsigned int maxSize = 0x2000000; /* 32 meg is the largest available */
	unsigned char crtcext3; /* save contents of mga reg */
	unsigned int detected;

#if LINUX_VERSION_CODE >= 0x020300
	mga->ctrl = ioremap_nocache(mga->dev->resource[1].start, 0x4000);
	mga->fb = ioremap(mga->dev->resource[0].start, maxSize);
#else
	mga->ctrl = ioremap_nocache(mga->dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK,0x4000);
	mga->fb = ioremap(mga->dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK, maxSize);
#endif

        maxSize &= ~0x1FFFFF;   /* must be X*2MB (really it must be 2 or X*4MB) */
        /* at least 2MB */
        if (maxSize < 0x0200000) return 0;
        if (maxSize > 0x2000000) maxSize = 0x2000000;

	writeb(0x03, mga->ctrl + CRTCEXTADD);
	crtcext3 = readb(mga->ctrl + CRTCEXTREG);               /* save the reg */
	writeb((crtcext3 | 0x80), mga->ctrl + CRTCEXTREG);      /* turn on power graphics mode */
	store = readb(mga->fb + 0x1234);
        tmp = bytes;

        for (offs = 0x100000; offs < maxSize; offs += 0x200000)
                *tmp++ = readb(mga->fb + offs);

        for (offs = 0x100000; offs < maxSize; offs += 0x200000)
                writeb(0x02, mga->fb + offs);
	writeb(0x00, mga->ctrl + CACHEFLUSH);

        for (offs = 0x100000; offs < maxSize; offs += 0x200000) {
                if (readb(mga->fb + offs) != 0x02)
                        break;
                writeb(readb(mga->fb + offs) - 0x02, mga->fb + offs);
                if (readb(mga->fb + offs))
                        break;
        }

        tmp = bytes;
        for (offs2 = 0x100000; offs2 < maxSize; offs2 += 0x200000)
                writeb(*tmp++, mga->fb + offs2);

        writeb(store, mga->fb + 0x1234);
	writeb(0x03, mga->ctrl + CRTCEXTADD);
	writeb(crtcext3, mga->ctrl + CRTCEXTREG);       /* restore prior reg value */

	detected = offs - 0x100000;
	printk("mgacore: %d MB detected\n", detected / 0x100000);

	iounmap(mga->ctrl);
	iounmap(mga->fb);

        return detected;
}
               
#ifndef PCI_DEVICE_ID_MATROX_G200_PCI 
#define PCI_DEVICE_ID_MATROX_G200_PCI 0x0520
#endif

#ifndef PCI_DEVICE_ID_MATROX_G200_AGP 
#define PCI_DEVICE_ID_MATROX_G200_AGP 0x0521
#endif

#ifndef PCI_DEVICE_ID_MATROX_G400 
#define PCI_DEVICE_ID_MATROX_G400 0x0525
#endif


static void mga_init_device(struct mga_dev_s *mga, struct pci_dev *dev)
{
	memset(mga, 0, sizeof(struct mga_dev_s));
  
	mga->dev = dev;
	mga->memory_size = mga_guess_memory_size(mga);

#if LINUX_VERSION_CODE >= 0x020300
	mga->ctrl = ioremap_nocache(dev->resource[1].start, 0x4000);
	mga->fb = ioremap(dev->resource[0].start, mga->memory_size);
#else
	mga->ctrl = ioremap_nocache(dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK, 0x4000);
	mga->fb = ioremap(dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK, mga->memory_size);
#endif
	/* we hit the mouse cursor if we start from the end of memory */
	mga->malloc_limit = mga->memory_size - 0x100000;
}

/* * PRIVATE FUNCTION *
 *	mga_detect_card -	detect a Matrox card
 *
 *	This function tries to detect a Matrox card (G series)
 *
 *	Return value:
 *	Returns 0 on success.
 *
 *	Returns %-ENODEV on failure.
 *
 *	Notes:	only detects G200 and G400 at this time. It should be
 *	easy to add other models to the list.
 */
static int mga_detect_card(void)
{
	struct pci_dev *dev;
	int i = 0;

	dev = NULL;
	while ((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G400, dev))) {
		printk("mgacore: Found MGA G400\n");
		mga_init_device(&card[i++], dev);
	}

	dev = NULL;
	while ((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G200_AGP, dev))) {
			printk("mgacore: Found MGA G200 AGP\n");
			mga_init_device(&card[i++], dev);
	}

	dev = NULL;
	while ((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G200_PCI, dev))) {
			printk("mgacore: Found MGA G200 PCI\n");
			mga_init_device(&card[i++], dev);
	}
	
	if (!card[0].ctrl) {
		printk("mgacore: No supported cards found\n");
		return -EIO;   
	}

	return 0;
}

/**
 *	mga_get_card -	return a device structure
 *
 *	This function returns a pointer to a &mga_dev_s device.
 */
struct mga_dev_s * mga_get_card(void)
{
	/* FIXME: how to handle multiple cards,
	 * and still be able to share one between
	 * two modules (like mgayuv and mgadvd)
	 */
	card[0].in_use++;
	MOD_INC_USE_COUNT;
	return &card[0];
}

/**
 *	mga_release_card -	reset the usage count of the card
 *	@mga:	mga instance
 *
 *	This function resets the usage counter of the device.
 */
void mga_release_card(struct mga_dev_s *mga)
{
	if (!mga)
		return;

	mga->in_use--;
	MOD_DEC_USE_COUNT;
}

int init_module(void)
{
	printk("Matrox core interface v0.1 (c) David Barth\n");

	return mga_detect_card();
}

void cleanup_module(void)
{
	int i;

	for (i=0; i<MAX_CARDS; i++) {
		if (card[i].ctrl) iounmap(card[i].ctrl);
		if (card[i].fb) iounmap(card[i].fb);
	}

}
