/*
 * mgayuv.c
 *
 * New version of mga_vid using mgacore
 *
 * Based on code Copyright (C) 1999 Aaron Holtzman
 *
 * This software has been released under the terms of the GNU Public
 * license. See http://www.gnu.org/copyleft/gpl.html for details.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/videodev.h>

#ifndef VIDEO_WINDOW_CHROMAKEY
#define VIDEO_WINDOW_CHROMAKEY	16
#endif /* VIDEO_WINDOW_CHROMAKEY */

#include "mgacore.h"
#include "mgabes.h"

#define MGAYUV_ID_STRING "Matrox G200/G400 YUV"

static struct mga_dev_s *mga;
static int norm = VIDEO_MODE_PAL;
struct video_window vw = { 0, 0, 720, 480 };
struct video_buffer vbuf = { 0, 720, 480};
static u32 frame_size = 0;

static long mgayuv_read(struct video_device *v, char *buf, unsigned long count, 
	int noblock)
{
	return -EINVAL;
}

static long mgayuv_write(struct video_device *v, const char *buf, unsigned long count, int noblock)
{
	return -EINVAL;
}

static int attach_buffer(struct mga_dev_s *mga)
{
	if (vbuf.base)
		mga_free_buffer(mga, frame_size);

	/* allocate a buffer for a video frame */
	vbuf.bytesperline = mgabes_pitch_rounding(mga->bes, vbuf.width);
	frame_size = vbuf.bytesperline * vbuf.height * 2; /* YUV420 */
	vbuf.base = mga_alloc_buffer(mga, frame_size);
	mgabes_set_source(mga->bes, vbuf.width, vbuf.height, (u32)vbuf.base,
			  0 /* no frame B */, VIDEO_PALETTE_YUV420);
	mgabes_set_frame(mga->bes, BES_FRAME_A);
	mgabes_set_window(mga->bes, vw.x, vw.y, vw.width, vw.height);

	return 0;
}

static int mgayuv_mmap(struct video_device *dev, const char *addr, unsigned long size)
{
	struct mga_dev_s *mga;
        unsigned long phys_addr;

	mga = (struct mga_dev_s *)dev->priv;
	if (!dev->priv) return -ENODEV;

#if LINUX_VERSION_CODE >= 0x020300
	phys_addr = mga->dev->resource[0].start + (u32)vbuf.base;
#else
	phys_addr = (mga->dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK) + (u32)vbuf.base;
#endif

	if (remap_page_range((unsigned long)addr, phys_addr, size, PAGE_SHARED)) {
		printk("mgayuv: error mapping video buffer\n");
		return -EAGAIN;
	}

	return 0;
}

static int mgayuv_open(struct video_device *dev, int mode)
{
	struct mga_dev_s *mga;

	mga = (struct mga_dev_s *)dev->priv;
	if (!mga) return -ENODEV;

	mgabes_init(mga);
	mgabes_set_mode(mga->bes, BES_HFILTER | BES_VFILTER | BES_ACC2XZOOM | BES_DITHER);
	attach_buffer(mga);

	MOD_INC_USE_COUNT;
	return 0;
}

static void mgayuv_close(struct video_device *dev)
{
	struct mga_dev_s *mga;

	mga = (struct mga_dev_s *)dev->priv;
	if (!mga) return;

	mgabes_desactivate(mga->bes);

	if (vbuf.base)
		mga_free_buffer(mga, frame_size);
	vbuf.base = 0;

	MOD_DEC_USE_COUNT;
	return;
}

static int mgayuv_init_done(struct video_device *dev)
{
	return 0;
}

static int mgayuv_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct mga_dev_s *mga;
	struct mga_bes_s *bes;

	mga = (struct mga_dev_s *)dev->priv;
	if (!mga) return -ENODEV;

	bes = mga->bes;

	switch(cmd)
	{
		case VIDIOCGCAP:
		{
			struct video_capability b;
			strcpy(b.name, MGAYUV_ID_STRING);
			b.type = VID_TYPE_SCALES | VID_TYPE_OVERLAY | VID_TYPE_CHROMAKEY;
			b.channels = 1;
			b.audios = 0;
			b.maxwidth = 1024;	/* FIXME: check the specs */
			b.maxheight = 768;
			b.minwidth = 32;
			b.minheight = 16;
			if(copy_to_user(arg,&b,sizeof(b)))
				return -EFAULT;
			return 0;
		}
                case VIDIOCGFBUF:
		{
			struct video_buffer fake_fb_info;
			/*
			 * In all this driver code we use the 'base'
			 * field as the relative address of the video
			 * buffer inside the framebuffer (for
			 * convenience), whereas the standard says
			 * it's the physical address.
			 */
			fake_fb_info.width = vbuf.width;
			fake_fb_info.height = vbuf.height;
			fake_fb_info.depth = vbuf.depth;
			fake_fb_info.bytesperline = vbuf.bytesperline;

#if LINUX_VERSION_CODE >= 0x020300
			fake_fb_info.base += mga->dev->resource[0].start;
#else
			fake_fb_info.base += (mga->dev->base_address[0] &
					      PCI_BASE_ADDRESS_MEM_MASK);
#endif
			if(copy_to_user(arg, &fake_fb_info, 
					sizeof(struct video_buffer)))
				return -EFAULT;
			return 0;
		}
                case VIDIOCSFBUF:
		{
			return -EINVAL;
		}
                case VIDIOCGCHAN:
		{
			struct video_channel v;
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if(v.channel != 0)
				return -EINVAL;
			v.flags = 0;
			v.tuners = 0;
			v.type = VIDEO_TYPE_CAMERA;
			v.norm = VIDEO_MODE_AUTO;
			strcpy(v.name, "Video Overlay");
			if(copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}
                case VIDIOCSCHAN:
		{
			struct video_channel v;
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if(v.channel != 0)
				return -EINVAL;
			if(v.norm == VIDEO_MODE_AUTO)
				return -EINVAL;

			/*
			 * change overlay buffer dimension
			 */
			if ((v.norm == VIDEO_MODE_PAL) ||
			    (v.norm == VIDEO_MODE_SECAM)) {
				vbuf.width = 768;
				vbuf.height = 576;
			} else { /* VIDEO_MODE_NTSC */
				vbuf.width = 640;
				vbuf.height = 480;
			}
			attach_buffer(mga);
			norm = v.norm;
			return 0;
		}

		case VIDIOCSWIN:
		{
			if(copy_from_user(&vw, arg, sizeof(vw)))
				return -EFAULT;
			if(vw.x <0 || vw.y <0 || vw.width < 32 
				|| vw.height < 16)
				return -EINVAL;
			if (vw.flags == VIDEO_WINDOW_CHROMAKEY)
				mgabes_set_colorkey(bes, vw.chromakey);
			mgabes_set_window(bes, vw.x, vw.y, vw.width, vw.height);

			return 0;
				
		}
		case VIDIOCGPICT:
		{
			/*
			 *	Default values.. if we can change this we
			 *	can add the feature later
			 */
			struct video_picture vp;
			vp.brightness = 0x8000;
			vp.hue = 0x8000;
			vp.colour = 0x8000;
			vp.whiteness = 0x8000;
			vp.depth = 8;
			vp.palette = VIDEO_PALETTE_YUV420;
			if(copy_to_user(arg, &vp, sizeof(vp)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCGWIN:
		{
			if(copy_to_user(arg, &vw, sizeof(struct video_window)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCSPICT:
		{
			return -EINVAL;
		}
		case VIDIOCCAPTURE:
		{
			if (arg)
				mgabes_activate(bes);
			else
				mgabes_desactivate(bes);

			return 0;
		}
		default:
			return -ENOIOCTLCMD;
	}
}

static struct video_device mgayuv_dev =
{
	MGAYUV_ID_STRING,
	VID_TYPE_OVERLAY | VID_TYPE_CHROMAKEY | VID_TYPE_SCALES,
	VID_HARDWARE_BT848,		/* This is a lie for now */
	mgayuv_open,
	mgayuv_close,
	mgayuv_read,
	mgayuv_write,
	NULL,
	mgayuv_ioctl,
	mgayuv_mmap,
	mgayuv_init_done,
	NULL,
	0,
	0
};




/* 
 * Main Initialization Function 
 */

static int mgayuv_initialize(void)
{
	printk(MGAYUV_ID_STRING"\n");

	if (!(mga = mga_get_card()))
		return -EINVAL;
	
	if (video_register_device(&mgayuv_dev, VFL_TYPE_GRABBER) < 0) {
		printk("mgayuv: unable to register.\n");
		mga_release_card(mga);
		return -EINVAL;
	}

	mgayuv_dev.priv = mga;

	return 0;
}

int init_module(void)
{
	return mgayuv_initialize();
}

void cleanup_module(void)
{
	video_unregister_device(&mgayuv_dev);
	mga_release_card(mga);
}
