#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#if defined(__NetBSD__) || defined(__OpenBSD__)
# include <sys/dvdio.h>
#elif defined(__linux__)
# include <linux/cdrom.h>
#else
# error "Need the DVD ioctls"
#endif

#include "css_auth.h"
#include "css.h"


typedef struct {
	u_char Challenge[10];
	u_char Key1[LEN_KEY];
	u_char Key2[LEN_KEY];
	u_char KeyCheck[LEN_KEY];
	u_char varient;
} DiscKeys;

DiscKeys discKeys;

#ifdef DEBUG
/**
 * Print Functions
 **/

static void _CSSPrintBytes (const u_char *val, int len)
{
	const u_char *ptr = val;

	for (; len > 0; len--)
		fprintf (stderr, " %02X", *ptr++ & 0xff);
}

#define _CSSPrintDebug(fmt, arg...) fprintf (stderr, fmt"\n", ##arg)
#define _CSSPrintDebugBytes(descr, val, len) fprintf (stderr, descr":\t"); _CSSPrintBytes (val, len); fprintf (stderr, "\n")
#else
#define _CSSPrintDebug(fmt, arg...)
#define _CSSPrintDebugBytes(descr, val, len)
#endif

/**
 * CSSAuthDrive
 * Determine if drive is authentic
 * Return: -1 if failed, 0..32 = varient of CSS
 **/


static int _CSSAuthDrive (const u_char *key, DiscKeys *disc)
{
	int i;

	for (i=0; i<LEN_KEY; i++)
		disc->Key1[i] = key[4-i];

	for (i=0; i<32; ++i) {
		CryptKey1( i, disc->Challenge, disc->KeyCheck);

		if (memcmp (disc->KeyCheck, disc->Key1, LEN_KEY) == 0) {
			_CSSPrintDebug ("Drive Authentic - using varient %d", i);
			disc->varient = i;
			return i;
		}
	}

	_CSSPrintDebug ("Drive would not Authenticate");

	return -1;
}


/**
 * Get Disk Key
 * Input:
 * fd 	device handle of DVD
 * agid
 * key 	player key
 * 
 * Output:
 * diskkey	2048 byte block of disk key XOR player key
 * return 	0=failed, 1 = success
 **/

static int _CSSGetDiskKey (int fd, int agid, char *key, char *key_disc)
{
	dvd_struct s;
	int i;

	s.type = DVD_STRUCT_DISCKEY;
	s.disckey.agid = agid;
	memset (s.disckey.value, 0, 2048);

	if (ioctl(fd, DVD_READ_STRUCT, &s)<0) {
		perror ("Could not read Disc Key");
		return -1;
	}

	for (i=0; i<sizeof s.disckey.value; i++)
		s.disckey.value[i] ^= key[4 - (i % LEN_KEY)];

	_CSSPrintDebugBytes ("Received Disc Key", s.disckey.value, 10);

	memcpy (key_disc, s.disckey.value, 2048);

	return 0;
}


/**
 * Get Title Key
 * Input:
 * fd 		device handle of DVD
 * agid
 * key 		player key
 * 
 * Output:
 * key_title	2048 byte block of disk key XOR player key
 * return 		0=failed, 1 = success
 **/

static int _get_title_key (int fd, int agid, int lba, char *key, char *key_title)
{
	dvd_authinfo ai;
	int i;

	ai.type = DVD_LU_SEND_TITLE_KEY;

	ai.lstk.agid = agid;
	ai.lstk.lba = lba;

	if (ioctl (fd, DVD_AUTH, &ai)) {
		perror ("GetTitleKey failed");
		return -1;
	}

	for (i=0; i< LEN_KEY; ++i)
		ai.lstk.title_key[i] ^= key[4 - (i%LEN_KEY)];

	_CSSPrintDebugBytes ("Received Title Key", ai.lstk.title_key, LEN_KEY);
	_CSSPrintDebug (" CPM=%d, CP_SEC=%d, CGMS=%d", ai.lstk.cpm, ai.lstk.cp_sec, ai.lstk.cgms);

	memcpy (key_title, ai.lstk.title_key, LEN_KEY);

	return 0;
}


/**
 * Get ASF (Authentication Success Flag)
 * fd 	device handle of DVD
 * Output:
 * return: 0 = failure, 1 = success
 **/

int CSSGetASF (int fd)
{
	dvd_authinfo ai;

	ai.type = DVD_LU_SEND_ASF;
	ai.lsasf.agid = 0;
	ai.lsasf.asf = 0;

	if (ioctl(fd, DVD_AUTH, &ai)) {
		perror ("GetASF");
		return -1;
	}

	_CSSPrintDebug ("%sAuthenticated", (ai.lsasf.asf) ? "" : "not ");

	return 0;
}


/**
 *
 **/

int _CSSGetAgid( int fd, dvd_authinfo *ai)
{
	int tries;
	int rv;

	/* Init sequence, request AGID */
	for (tries = 1; tries < 4; ++tries) {
		_CSSPrintDebug ("Request AGID [%d]...\t", tries);
		ai->type = DVD_LU_SEND_AGID;
		ai->lsa.agid = 0;
		rv =  ioctl (fd, DVD_AUTH, ai);
		if (rv != -1) {
			// found key
			return( tries);
		}
		perror ("N/A, invalidating");
		ai->type = DVD_INVALIDATE_AGID;
		ai->lsa.agid = 0;
		ioctl (fd, DVD_AUTH, ai);
	}

	_CSSPrintDebug ("Cannot get AGID");

	return -1;
}


/**
 * Simulation of a non-CSS compliant host (i.e. the authentication fails,
 * but idea is here for a real CSS compliant authentication scheme).
 **/

static int _CSSAuthHost (dvd_authinfo *ai, DiscKeys *disc)
{
	int i;

	switch (ai->type) {
	/* Host data receive (host changes state) */
	case DVD_LU_SEND_AGID:
		_CSSPrintDebug ("AGID %d", ai->lsa.agid);

		ai->type = DVD_HOST_SEND_CHALLENGE;
		break;

	case DVD_LU_SEND_KEY1:
		_CSSPrintDebugBytes ("LU sent key1", ai->lsk.key, LEN_KEY);

		if (_CSSAuthDrive (ai->lsk.key, disc)<0) {
			ai->type = DVD_AUTH_FAILURE;
			return -EINVAL;
		}
		ai->type = DVD_LU_SEND_CHALLENGE;
		break;

	case DVD_LU_SEND_CHALLENGE:
		for (i = 0; i < 10; ++i)
			disc->Challenge[i] = ai->hsc.chal[9-i];

		_CSSPrintDebugBytes ("LU sent challenge", ai->hsc.chal, 10);

		CryptKey2 (disc->varient, disc->Challenge, disc->Key2);
		ai->type = DVD_HOST_SEND_KEY2;
		break;

	/* Host data send */
	case DVD_HOST_SEND_CHALLENGE:
		for (i = 0; i < 10; ++i)
			ai->hsc.chal[9-i] = disc->Challenge[i];

		_CSSPrintDebugBytes ("Host sending challenge", ai->hsc.chal, 10);

		/* Returning data, let LU change state */
		break;

	case DVD_HOST_SEND_KEY2:
		for (i = 0; i < LEN_KEY; ++i)
			ai->hsk.key[4-i] = disc->Key2[i];

		_CSSPrintDebugBytes ("Host sending key 2", &disc->Key2[0], LEN_KEY);

		/* Returning data, let LU change state */
		break;

	default:
		_CSSPrintDebug ("Got invalid state %d", ai->type);

		return -EINVAL;
	}

	return 0;
}


/**
 *
 **/

int CSSAuth (int fd, DiscKeys *disc)
{
 	dvd_authinfo ai;
	int i, agid;

	memset (&ai, 0, sizeof (ai));

	if (CSSGetASF (fd))
		return -1;

	_CSSGetAgid (fd, &ai);

	for (i = 0; i < 10; ++i)
		disc->Challenge[i] = i;

	/* Send AGID to host */
	if (_CSSAuthHost(&ai, disc) < 0) {
		_CSSPrintDebug ("Send AGID to host failed");
		return -1;
	}
	/* Get challenge from host */
	if (_CSSAuthHost(&ai, disc) < 0) {
		_CSSPrintDebug ("Get challenge from host failed");
		return -1;
	}
	agid = ai.lsa.agid;
	/* Send challenge to LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Send challenge to LU failed");
		return -1;
	}
	/* Get key1 from LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Get key1 from LU failed");
		return -1;
	}
	/* Send key1 to host */
	if (_CSSAuthHost(&ai, disc) < 0) {
		_CSSPrintDebug ("Send key1 to host failed");
		return -1;
	}
	/* Get challenge from LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Get challenge from LU failed");
		return -1;
	}
	/* Send challenge to host */
	if (_CSSAuthHost (&ai, disc) < 0) {
		_CSSPrintDebug ("Send challenge to host failed");
		return -1;
	}
	/* Get key2 from host */
	if (_CSSAuthHost (&ai, disc) < 0) {
		_CSSPrintDebug ("Get key2 from host failed");
		return -1;
	}
	/* Send key2 to LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Send key2 to LU failed (expected)");
		return -1;
	}

	if (ai.type == DVD_AUTH_ESTABLISHED)
		_CSSPrintDebug ("DVD is authenticated");
	else if (ai.type == DVD_AUTH_FAILURE)
		_CSSPrintDebug ("DVD authentication failed");

	memcpy (disc->Challenge, disc->Key1, LEN_KEY);
	memcpy (disc->Challenge+LEN_KEY, disc->Key2, LEN_KEY);
	CryptBusKey (disc->varient, disc->Challenge, disc->KeyCheck);

	_CSSPrintDebugBytes ("Received Session Key", &disc->KeyCheck[0], LEN_KEY);

	return( agid);

}


/**
 * CSSAuthDisk: Authorise disk
 * fd	device
 * key	player key
 **/
	
int CSSAuthDisk (int fd, char *key)
{
	int agid;
	DiscKeys *disc = &discKeys;

	agid = CSSAuth( fd, disc);

	if (agid < 0)
		return -1;

	if (CSSGetASF (fd) < 0)
		return -1;

	_CSSGetDiskKey (fd, agid, disc->KeyCheck, key);

	if (CSSGetASF (fd) < 0)
		return -1;

	return 0;
}


/** 
 * CSSAuthTitle
 * Authorise title from disk
 * fd	device
 * key	disc key
 * lba	logical block address of title	
 **/

int CSSAuthTitle (int fd, char *key, int lba)
{
	int agid;
	DiscKeys *disc = &discKeys;

	agid = CSSAuth( fd, disc);
	if (agid < 0) return -1;

	_get_title_key (fd, agid, lba, disc->KeyCheck, key);

	if (CSSGetASF (fd) < 0)
		return -1;

	return 0;
}
