
/*
 *
 * Copyright (C) 1999  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 * May 29th 2000: Dave Sainty <dave@dtsp.co.nz>
 *	NetBSD fixes
 *	correced under/overflows (strncat)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>

#include "dvddb_conf.h"

struct server_struct {
	const char *name;
	int port;
};

#ifndef INADDR_NONE
#define INADDR_NONE 0xFFFFFFFF
#endif

static FILE *_dvddbOpen (const char *id, const char *file_mode);


/**
 * Open a connection to the DVDDB server.
 **/
	
int dvddbNetConnect (void)
{
	int sock;
	struct sockaddr_in sin;
	struct hostent *host;
	struct server_struct server;

	server.name = DVDDB_HOST;
	server.port = DVDDB_HOST_PORT;

	sin.sin_family = AF_INET;
	sin.sin_port = htons(server.port);

	if((sin.sin_addr.s_addr = inet_addr(server.name)) == INADDR_NONE) {
		if((host = gethostbyname(server.name)) == NULL) {
			perror ("gethostbyname");
			return -1;
		}

		memcpy(&sin.sin_addr, host->h_addr, host->h_length);
	}

	if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror ("socket");
		return -1;
	}

	if(connect(sock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		perror ("connect");
		return -1;
	}

	return sock;
}


/**
 * Close connection previously established to the DVDDB server.
 */

int dvddbNetDisconnect (int sock)
{
	return close (sock);
}


/**
 * Put a new entry in the database (on server) 
 */

int dvddbNetPut (int sock, int fd, char *id)
{
	char buf[BUFSIZ];
	int len;
#ifdef DEBUG
	int i;
#endif

// get file size
	len = (int) lseek (fd, 0, SEEK_END);
	lseek (fd, 0, SEEK_SET);
	
// form HTTP header
	if ((snprintf (buf, BUFSIZ, "POST /%s?%s HTTP/1.0\nContent-length: %d\n\n",
			DVDDB_HOST_DB,
			id,
			len)) == -1) {
		perror ("dvddbNetPut");
		return -1;
	}

// write HTTP header
	write (sock, buf, strlen (buf));

// write data
	while ((len = read (fd, buf, BUFSIZ)) > 0) {
#ifdef DEBUG
		for (i=0; i<len; i++)
			fprintf (stderr, "%c", buf[i]);
#endif
		if (write (sock, buf, len) < 0)
			return -1;
	}
	
	return 0;
}


/**
 * Get an entry from the server database (and cache it locally)
 * DENT: clean up here ... so this is a place where most likely 
 *	a seqfault appears	
 */

int dvddbNetGet (int sock, char *id)
{
	int http_hdr_found = 0;
	int db_hdr_found = 0;
	char buf[BUFSIZ];
	int len;
	int i;
	FILE *file = NULL;

// request specific ID from host
	if ((snprintf (buf, BUFSIZ, "GET /%s?%s HTTP/1.0\n\n",
			DVDDB_HOST_DB,
			id)) < 0) {
		fprintf (stderr, "dvddbNetPut: output truncated");
		return -1;
	}

	write (sock, buf, strlen (buf));

// and now, wait for reply, drop HTTP header and put rest int
//	local cache (file)
 
	while ((len = read (sock, buf, BUFSIZ)) > 0) {
		i = 0;

		if (!http_hdr_found) {
			for (i=0; (i<len-1) && !http_hdr_found; i++) {
				if (	(buf[i  ] == '\n') &&
			    		(buf[i+1] == '\r')) {

					http_hdr_found = 1;	
					i+=2;
				}
			}
		}

// parsing dvddb_hdr
		if (http_hdr_found && !db_hdr_found) {
			char const *token;
			token = strtok (buf+i, ":");

			while (!db_hdr_found && token) {
				if (!strcasecmp (token, "version")) {
					int version;

					version = atoi (strtok (NULL, "\n\r"));
				
					if (version > DVDDB_VERSION) {
						fprintf (stderr, "dvddbNetGet: server version (v%d) is greater than client version (v%d)\n - please upgrade\n", version, DVDDB_VERSION );
						fclose (file);
						return -1;
					}	
				} else if (!strcasecmp (token, "data")) {
					db_hdr_found = 1;
					token = strtok (NULL, "\n\r");
					if (!strcasecmp (token, " no"))
						return -1;
					token = strtok (NULL, "\n\r");
					i = token - &buf[0];

					if ((file = _dvddbOpen (id, "w")) == NULL)
					return -1;

					break;
				}

				token = strtok (NULL, ":");
			}
		} 

		if (db_hdr_found) {
			fwrite (&buf[i], len-i, 1, file);	
#ifdef DEBUG
			for (; i<len; i++)
				fprintf (stderr, "%c", buf[i]);
#endif
		}
	}

	fclose (file);

	return 0;
}


/**
 *
 */

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

static FILE *_dvddbOpen (char const *id, char const *file_mode)
{
	FILE *file;
	char oms_dir_name[MAXPATHLEN];
	char dvddb_dir_name[MAXPATHLEN];
	char file_name[MAXPATHLEN];
        struct stat stat_buffer;
	int index=0;
      
// check for oms user-config dir 
	if (snprintf (oms_dir_name, sizeof(oms_dir_name), "%s/.oms",
				getenv ("HOME") ? getenv ("HOME") : "") < 0)
		goto err;

        if (stat (oms_dir_name, &stat_buffer))
		mkdir (oms_dir_name, S_IRUSR | S_IWUSR | S_IXUSR);
  
// check for dvddb dir 
	if (snprintf (dvddb_dir_name, sizeof(dvddb_dir_name), "%s/dvddb",
				oms_dir_name) < 0)
		goto err;
   
        if (stat (dvddb_dir_name, &stat_buffer))
		mkdir (dvddb_dir_name, S_IRUSR | S_IWUSR | S_IXUSR);
 
// prepare to read/write from/to the id file   
	if (snprintf (file_name, sizeof(file_name), "%s/%s",
				dvddb_dir_name, id) < 0)
		goto err;

        fprintf (stderr, "_dvddbOpen %s for %s ...",
		file_name,
		*file_mode == 'r' ? "reading" : "writing");

	if (!(file = fopen (file_name, file_mode))) {
		fprintf (stderr, "dvddbOpen: cannot open DVDDB cache file (%s)\n", file_name);
                return NULL;
        }

	fprintf (stderr, "OK.\n");

        return file;

err:
	fprintf (stderr, "dvddbOpen: error creating DVDDB filename-string\n");
	return NULL;
}


/**
 *
 */

FILE *dvddbOpen (char *id)
{
	return _dvddbOpen (id, "r");
}

/**
 *
 */

int dvddbClose (FILE *file)
{
	if (!file)
		return -1;

	return fclose (file);
}


/**
 *
 */

char *dvddbRead (FILE *file)
{
	static char buf[BUFSIZ];

	if (!file)
		return NULL;

	if (fgets (buf, BUFSIZ, file)) {
		buf[strlen(buf)-1] = '\0';      // delete newline
		return buf;
	}

	return NULL;
}


/**
 * DENT: move this into the UDF code - if christian agrees
 *	SOMEONE SHOULD CHECK IF THIS ID IS UNIQUE
 */



#define DVD_BLKSIZE ((off_t) 0x800)
#define OFF_ID_BLK ((off_t) 0x20)
#define OFF_ID_BYTE ((off_t) 0x49)

int dvddbGetID (char *device, char *id)
{
	int fd;
	char buf[DVD_BLKSIZE];
	int ret = -1;

	if ((fd = open (device, O_RDONLY)) <= -1) {
		fprintf (stderr, "dvddbGetID: unable to open device (%s)", device);
		goto err;
	}

	if (lseek (fd, DVD_BLKSIZE * OFF_ID_BLK, SEEK_SET) <= -1) {
		perror ("lseek");
		goto err;
	}

	if (read (fd, buf, DVD_BLKSIZE) < DVD_BLKSIZE) {
		fprintf (stderr, "dvddbGetID: short read on ID\n");
		goto err;
	}

	memcpy (id, &buf[OFF_ID_BYTE], 8);
	id[8] = '\0';

	ret = 0;
err:
        close (fd);

        return ret;
}
