
/*
 * PROGRAM STREAM DECAPSULATION
 *
 * Copyright (C) 1999-2000  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include <dvb/ps.h>
#include <dvb/pes.h>
#include <oms/plugin/decaps.h>

static int _ps_open (plugin_decaps_t *plugin);
static int _ps_close (plugin_decaps_t *plugin);
static int _ps_decaps				(buf_t *buf);
static plugin_decaps_magic_t *_ps_magic		(void);
static int _ps_set_filter			(u_int filter_id, int val);

static struct ps_priv_struct {
	u_int filter[_BUF_MAX];
} priv;

static plugin_decaps_t decaps = {
	_ps_open,
	_ps_close,
	_ps_decaps,	// decasulation
	_ps_magic,	// autodetect decapsulation
	_ps_set_filter,	// filtering
	NULL,		// read (inserted)
	NULL,		// write (inserted)
	2048,		// blocksize (inserted: this is a default)
	NULL		// private data
};

static buf_t *buf; 
static buf_entry_t buf_entry; 

#define AUDIO_LPCM	0xA0
#define AUDIO_AC3	0x80

#define PS_FILTER_VIDEO	0xE0
#define PS_FILTER_SPU	0x20

static char ps_file_magic[] = {
	0x00, 0x00, 0x01, 0xBA
};

static plugin_decaps_magic_t ps_magic_entry = {
	"ps;vob",
	4,
	ps_file_magic
};


/**
 *
 **/

static int _ps_open (plugin_decaps_t *plugin)
{
	buf_entry.mem_len = decaps.blocksize;

	priv.filter[BUF_VIDEO] = PS_FILTER_VIDEO;
	priv.filter[BUF_AUDIO] = 0x00;	// audio substream id (0...7, -1=off)
	priv.filter[BUF_SUBPIC] = 0x20;	// subtitle substream id (0..31,-1=off)

	return 0;
}

/**
 *
 **/

static int _ps_close (plugin_decaps_t *plugin)
{
	return 0;
}

/**
 *
 **/

static inline char *_get_bytes_block (u_int len)
{
	if (!buf_entry.mem) {
		buf_entry.mem = valloc (decaps.blocksize);// check return val
		buf_entry.mem_len = decaps.blocksize;
	}

	if (decaps.read (buf, &buf_entry) <= 0)
		return NULL;

	buf_entry.data = buf_entry.mem + len;

	return buf_entry.mem;
}

/**
 *
 */

static inline char *_get_bytes (u_int len)
{
	char *ptr = buf_entry.data;
	buf_entry.data += len;

	return ptr;
//	return buf_entry.data+=len;
}


/**
 *
 */

static int _ps_decaps (buf_t *_buf)
{
	ps_hdr_t *ps;
	pes_hdr_t *pes;
	pes_flag_t *flag;

	buf = _buf;

//get new block
	if (!(ps = (ps_hdr_t *) _get_bytes_block (PS_HDR_LEN)))
		return -1;

	buf_entry.flags = BUF_FLAG_NONE;

// PS
	if (ps->stream_id == PS_STREAM_PACK_HDR) {
		ps_pack_t *ps_pack = (ps_pack_t *) _get_bytes (PS_PACK_LEN);

		buf_entry.clock = (ps_pack->system_clock_reference_1 << 30) |
				(ps_pack->system_clock_reference_2 << 28) |
				(ps_pack->system_clock_reference_3 << 20) |
				(ps_pack->system_clock_reference_4 << 15) |
				(ps_pack->system_clock_reference_5 << 13) |
				(ps_pack->system_clock_reference_6 << 5) |
				(ps_pack->system_clock_reference_7);
	
		buf_entry.clock_ext = (ps_pack->system_clock_reference_ext_1 << 7) |
				ps_pack->system_clock_reference_ext_2;

		buf_entry.flags |= BUF_FLAG_CLOCK_VALID;

		_get_bytes (ps_pack->pack_stuffing_length);

		ps = (ps_hdr_t *) _get_bytes (PS_HDR_LEN);
	}

	if (ps->stream_id == PS_STREAM_SYSTEM_HDR) {
		ps_system_hdr_t *ps_system = (ps_system_hdr_t *) _get_bytes (2);
		u_int len;

		len = (ps_system->header_length_hi << 8) |
			ps_system->header_length_lo;

		_get_bytes (len);

		ps = (ps_hdr_t *) _get_bytes (PS_HDR_LEN);
	}

// PES
	buf_entry.data -= PS_HDR_LEN;
	pes = (pes_hdr_t *) _get_bytes (PES_HDR_LEN);
	flag = (pes_flag_t *) _get_bytes (PES_FLAG_LEN);

	buf_entry.data_len = ((pes->pes_packet_length_hi << 8) |
		pes->pes_packet_length_lo) -
		PES_FLAG_LEN -
		flag->pes_header_data_length;

	_get_bytes (flag->pes_header_data_length);

	if (flag->pts_dts_flag) {
		u_int8_t *ptr = (u_int8_t *) (((u_char *) (flag)) + PES_FLAG_LEN);
		buf_entry.pts  = (*ptr++ & 0x0E) << 29;	// >> 1 << 30
		buf_entry.pts |= (*ptr++       ) << 22;
		buf_entry.pts |= (*ptr++ & 0xFE) << 14; // >> 1 << 15
		buf_entry.pts |= (*ptr++       ) <<  7;
		buf_entry.pts |= (*ptr++ & 0xFE) >>  1;

		buf_entry.flags |= BUF_FLAG_PTS_VALID;
	}

	if (pes->stream_id == priv.filter[BUF_VIDEO]) {	// Video Stream
		buf_entry.buf_id = BUF_VIDEO;
	} else if (pes->stream_id == 0xBD) {		// Audio Stream, SPU
		if (*buf_entry.data == priv.filter[BUF_SUBPIC]) { // SPU
			_get_bytes (1);	
			buf_entry.data_len -= 1;
			buf_entry.buf_id = BUF_SUBPIC;
		} else {				// AUDIO	
			if ((*buf_entry.data & 0x0F) != priv.filter[BUF_AUDIO])
				return 0; // filter == -1 || stream != filter

			switch (*buf_entry.data & 0xF0) {
			case AUDIO_AC3:
				_get_bytes (4);	
				buf_entry.data_len -= 4;
				buf_entry.type = BUF_TYPE_AC3;
				break;
			case AUDIO_LPCM: 
				_get_bytes (7);	
				buf_entry.data_len -= 7;
				buf_entry.type = BUF_TYPE_LPCM;
				break;
			default:
				//fprintf(stderr,"unhandled audio stream: %x\n",*buf_entry.data);
				return 0;
			}

			buf_entry.buf_id = BUF_AUDIO;
		}
	} else {			// Unhandled Stream
//		fprintf (stderr, "unhandled stream_id %x\n", ps->stream_id);

		return 0;
	}

	decaps.write (buf, &buf_entry);
	buf_entry.mem = NULL;

	return 0;
}


/**
 * Autodetect decapsulation method
 */

static plugin_decaps_magic_t *_ps_magic (void)
{
	return &ps_magic_entry;
}


/**
 * Filter streams
 */

static int _ps_set_filter (u_int filter_id, int val)
{
	if (filter_id > _BUF_MAX)
		return -1;

	if (val < 0)
		val = 0xffff;
	else
		switch (filter_id) {
		case BUF_VIDEO:
			val += PS_FILTER_VIDEO;
			break;
		case BUF_AUDIO:
			break;
		case BUF_SUBPIC:
			val += PS_FILTER_SPU;
			break;
		}

	priv.filter[filter_id] = val;

	return 0;
}


/**
 * Initial condition for this nice little thingy
 */

void *plugin_init (char *whoami)
{
	pluginRegister (whoami,
		PLUGIN_ID_DECAPS,
		STREAM_ID_MPEG2,
		&decaps);

//FIXME
	_ps_open (&decaps);

	return &decaps;
}


/**
 * Drop it
 */

void plugin_exit (void)
{
}

